@@ -1,23 +1,23 @@
 """AsusRouter module"""
 
+from __future__ import annotations
+
 import logging
 import asyncio
 from datetime import datetime
 
-from asusrouter.connection import Connection
-import asusrouter.helpers as helpers
+from asusrouter import AsusRouterError, Connection, helpers
+from .const import(
+    AR_PATH,
+    DEFAULT_SLEEP_TIME,
+    MSG_ERROR,
+    MSG_INFO,
+    MSG_SUCCESS,
+    MSG_WARNING,
+)
 
 _LOGGER = logging.getLogger(__name__)
 
-DEFAULT_SLEEP_TIME = 0.1
-
-_MSG_SUCCESS_COMMAND = "Command was sent successfully"
-_MSG_SUCCESS_HOOK = "Hook was sent successfully"
-_MSG_SUCCESS_LOAD = "Page was loaded successfully"
-
-_MSG_ERROR_NO_CONTROL = "Device is connected in no-control mode. Sending commands is blocked"
-_MSG_ERROR_NO_MONITOR = "Device is connected in no-control mode. Sending hooks is blocked"
-
 INTERFACE_TYPE = {
     "wan_ifnames": "wan",
     "wl_ifnames": "wlan",
@@ -236,8 +236,8 @@ MONITOR_MAIN = {
 }
 
 TRAFFIC_GROUPS = {
-    "INTERNET" : "WAN0",    # main WAN
-    "INTERNET1" : "WAN1",   # secondary WAN (USB modem / phone)
+    "INTERNET" : "WAN",    # main WAN
+    "INTERNET1" : "USB",   # secondary WAN (USB modem / phone)
     "WIRED" : "WIRED",      # wired connections
     "BRIDGE" : "BRIDGE",    # bridge
     "WIRELESS0" : "WLAN0",  # 2.4 GHz WiFi
@@ -282,7 +282,7 @@ class AsusRouter:
 
         self._device_cpu_cores : list | None = None
         self._device_ports : dict | None = None
-        self._device_boottime: datetime | None = None
+        self._device_boottime : datetime | None = None
 
 
         self._monitor_main : dict | None = None
@@ -294,6 +294,9 @@ class AsusRouter:
         self._monitor_misc : dict | None = None
         self._monitor_misc_time : datetime | None = None
         self._monitor_misc_running : bool = False
+        self._monitor_devices : dict | None = None
+        self._monitor_devices_time : datetime | None = None
+        self._monitor_devices_running : bool = False
 
 
         """Connect"""
@@ -332,18 +335,23 @@ class AsusRouter:
         result = {}
 
         if not self._enable_monitor:
-            _LOGGER.error(_MSG_ERROR_NO_MONITOR)
+            _LOGGER.error(MSG_ERROR["disabled_monitor"])
             return result
 
         if hook is None:
             return result
 
         try:
-            result = await self.connection.async_run_command("hook={}".format(hook))
-            _LOGGER.debug("{}: {}".format(_MSG_SUCCESS_HOOK, hook))
+            result = await self.connection.async_run_command(command = "hook={}".format(hook))
+            _LOGGER.debug("{}: {}".format(MSG_SUCCESS["hook"], hook))
         except Exception as ex:
             _LOGGER.error(ex)
 
+        # Check for errors during hook
+        if self.connection.error:
+            _LOGGER.debug("Error flag found")
+            await self.async_handle_error()
+
         return result
 
 
@@ -353,7 +361,7 @@ class AsusRouter:
         result = {}
 
         if not self._enable_control:
-            _LOGGER.error(_MSG_ERROR_NO_CONTROL)
+            _LOGGER.error(MSG_ERROR["disabled_control"])
             return result
 
         request : dict = {
@@ -363,8 +371,8 @@ class AsusRouter:
             request[command] = commands[command]
 
         try:
-            result = await self.connection.async_run_command(str(request), "applyapp.cgi")
-            _LOGGER.debug("{}: {}".format(_MSG_SUCCESS_COMMAND, request))
+            result = await self.connection.async_run_command(command = str(request), endpoint = AR_PATH["command"])
+            _LOGGER.debug("{}: {}".format(MSG_SUCCESS["command"], request))
         except Exception as ex:
             _LOGGER.error(ex)
 
@@ -377,7 +385,7 @@ class AsusRouter:
         result = {}
 
         if not self._enable_monitor:
-            _LOGGER.error(_MSG_ERROR_NO_MONITOR)
+            _LOGGER.error(MSG_ERROR["disabled_monitor"])
             return result
 
         if page is None:
@@ -385,18 +393,16 @@ class AsusRouter:
 
         try:
             result = await self.connection.async_run_command(command = "", endpoint = page)
-            _LOGGER.debug(_MSG_SUCCESS_LOAD)
+            _LOGGER.debug(MSG_SUCCESS["load"])
         except Exception as ex:
             _LOGGER.error(ex)
 
         return result
 
-    
+
     async def async_monitor_main(self) -> None:
         """Get all the main monitor values. Non-cacheable"""
 
-        now = datetime.utcnow()
-
         while self._monitor_main_running:
             await asyncio.sleep(DEFAULT_SLEEP_TIME)
             return
@@ -408,6 +414,8 @@ class AsusRouter:
         hook = await self.async_compile_hook(MONITOR_MAIN)
         data = await self.async_hook(hook)
 
+        now = datetime.utcnow()
+
         ### CPU ###
         ## Not yet known what exactly is type of data. But this is the correct way to calculate usage from it
 
@@ -476,7 +484,11 @@ class AsusRouter:
             for el in monitor_main["NETWORK"]:
                 for nd in NETWORK_DATA:
                     if el in self._monitor_main["NETWORK"]:
-                        monitor_main["NETWORK"][el]["{}_speed".format(nd)] = (monitor_main["NETWORK"][el][nd] - self._monitor_main["NETWORK"][el][nd]) * 8
+                        traffic_diff = monitor_main["NETWORK"][el][nd] - self._monitor_main["NETWORK"][el][nd]
+                        # Handle overflow in the traffic stats, so there will not be a negative speed once per 0xFFFFFFFF + 1 bytes of data
+                        if (traffic_diff < 0):
+                            traffic_diff += 4294967296
+                        monitor_main["NETWORK"][el]["{}_speed".format(nd)] = traffic_diff * 8 / (now - self._monitor_main_time).total_seconds()
                     else:
                         monitor_main["NETWORK"][el]["{}_speed".format(nd)] = 0
 
@@ -491,8 +503,6 @@ class AsusRouter:
     async def async_monitor_nvram(self, groups : list[str] | str | None = None) -> None:
         """Get the NVRAM values for the specified group list"""
 
-        now = datetime.utcnow()
-
         while self._monitor_nvram_running:
             await asyncio.sleep(DEFAULT_SLEEP_TIME)
             return
@@ -520,6 +530,9 @@ class AsusRouter:
 
         message = await self.async_compile_nvram(requests)
         result = await self.async_hook(message)
+
+        now = datetime.utcnow()
+
         for item in result:
             monitor_nvram[item] = result[item]
 
@@ -539,8 +552,6 @@ class AsusRouter:
     async def async_monitor_misc(self) -> None:
         """Get all other useful values"""
 
-        now = datetime.utcnow()
-
         while self._monitor_misc_running:
             await asyncio.sleep(DEFAULT_SLEEP_TIME)
             return
@@ -555,7 +566,9 @@ class AsusRouter:
         ### PORTS
 
         # Receive ports number and status (disconnected, 100 Mb/s, 1 Gb/s)
-        data = await self.async_load("ajax_ethernet_ports.asp")
+        data = await self.async_load(page = AR_PATH["ports"])
+
+        now = datetime.utcnow()
 
         if "portSpeed" in data:
             monitor_misc["PORTS"] = dict()
@@ -569,7 +582,7 @@ class AsusRouter:
                         break
 
         # Load devicemap
-        data = await self.async_load("ajax_status.xml")
+        data = await self.async_load(page = AR_PATH["devicemap"])
         monitor_misc["DEVICEMAP"] = data
 
         # Calculate boot time. Since precision is 1 second, could be that old and new are 1 sec different. In this case, we should not change the boot time, but keep the previous value to avoid regular changes
@@ -583,9 +596,13 @@ class AsusRouter:
                 if "BOOTTIME" in self._monitor_misc:
                     _timestamp = self._monitor_misc["BOOTTIME"]["timestamp"]
                     _time = datetime.fromtimestamp(_timestamp)
-                    if time == _time:
+                    if (time == _time
+                        or abs(timestamp - _timestamp) < 2
+                    ):
                         monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
-                    elif abs(timestamp - _timestamp) < 2:
+                    # This happens on reboots if data is checked before the device got correct time from NTP
+                    elif timestamp - _timestamp < 0:
+                        # Leave the same boot time, since we don't know the new correct time
                         monitor_misc["BOOTTIME"] = self._monitor_misc["BOOTTIME"]
                     else:
                         monitor_misc["BOOTTIME"]["timestamp"] = timestamp
@@ -594,7 +611,8 @@ class AsusRouter:
                     monitor_misc["BOOTTIME"]["timestamp"] = timestamp
                     monitor_misc["BOOTTIME"]["ISO"] = time.isoformat()
 
-                self._device_boottime = time
+                if self._device_boottime != monitor_misc["BOOTTIME"]["ISO"]:
+                    self._device_boottime = monitor_misc["BOOTTIME"]["ISO"]
 
         self._monitor_misc = monitor_misc
 
@@ -604,6 +622,144 @@ class AsusRouter:
         return
 
 
+    async def async_monitor_devices(self) -> None:
+        """Monitor connected devices"""
+
+        while self._monitor_devices_running:
+            await asyncio.sleep(DEFAULT_SLEEP_TIME)
+            return
+
+        self._monitor_devices_running = True
+
+        if self._monitor_devices is None:
+            self._monitor_devices = dict()
+
+        monitor_devices = {}
+
+        data = (await self.async_hook("get_clientlist()"))["get_clientlist"]
+
+        now = datetime.utcnow()
+
+        if "maclist" in data:
+            for mac in data["maclist"]:
+                monitor_devices[mac] = dict()
+
+                # Name the device properly. If user has selected nickName manualy - it is the expected value. Else - name device set by itself or MAC address
+                if (data[mac]["nickName"] is not None
+                    and data[mac]["nickName"] != ""):
+                    monitor_devices[mac]["name"] = data[mac]["nickName"]
+                elif (data[mac]["name"] is not None
+                    and data[mac]["name"] != ""):
+                    monitor_devices[mac]["name"] = data[mac]["name"]
+                else:
+                    monitor_devices[mac]["name"] = mac
+
+                # Let's still include MAC address
+                monitor_devices[mac]["MAC"] = mac
+
+                # IP address
+                monitor_devices[mac]["IP"] = data[mac]["ip"]
+                # IP type
+                monitor_devices[mac]["IPMethod"] = data[mac]["ipMethod"]
+
+                # Internet info map
+                monitor_devices[mac]["internet"] = dict()
+                # Connected
+                if data[mac]["internetState"] is not None:
+                    monitor_devices[mac]["internet"]["online"] = data[mac]["internetState"]
+                # Mode
+                if data[mac]["internetMode"] is not None:
+                    monitor_devices[mac]["internet"]["mode"] = data[mac]["internetMode"]
+
+                # Connection info map
+                monitor_devices[mac]["connection"] = dict()
+
+                # Type of the connection
+                if data[mac]["isWL"] is not None:
+                    value = int(data[mac]["isWL"])
+                    if value == 1:
+                        monitor_devices[mac]["connection"]["type"] = "WLAN0"
+                    elif value == 2:
+                        monitor_devices[mac]["connection"]["type"] = "WLAN1"
+                    else:
+                        monitor_devices[mac]["connection"]["type"] = "LAN"
+                # Connected
+                if data[mac]["isOnline"] is not None:
+                    monitor_devices[mac]["connection"]["online"] = data[mac]["isOnline"]
+                    
+                # Only for wireless
+                if monitor_devices[mac]["connection"]["type"] != "LAN":
+                    # RSSI
+                    if data[mac]["rssi"] is not None:
+                        monitor_devices[mac]["connection"]["RSSI"] = data[mac]["rssi"]
+                    # Connection time
+                    if data[mac]["wlConnectTime"] is not None:
+                        value = await helpers.async_transform_connection_time(data[mac]["wlConnectTime"])
+                        timestamp = value.timestamp()
+                        iso = value.isoformat()
+
+                        if (mac in self._monitor_devices
+                            and "connection" in self._monitor_devices[mac]
+                            and "timestamp" in self._monitor_devices[mac]["connection"]
+                        ):
+                            _timestamp = self._monitor_devices[mac]["connection"]["timestamp"]
+                            if (timestamp == _timestamp
+                                or abs(timestamp - _timestamp) < 2
+                            ):
+                                monitor_devices[mac]["connection"]["timestamp"] = self._monitor_devices[mac]["connection"]["timestamp"]
+                                monitor_devices[mac]["connection"]["ISO"] = self._monitor_devices[mac]["connection"]["ISO"]
+                            else:
+                                monitor_devices[mac]["connection"]["timestamp"] = timestamp
+                                monitor_devices[mac]["connection"]["ISO"] = iso
+                        else:
+                            monitor_devices[mac]["connection"]["timestamp"] = timestamp
+                            monitor_devices[mac]["connection"]["ISO"] = iso
+
+                    # Connection speeds
+                    monitor_devices[mac]["connection"]["speed"] = dict()
+                    raw = data[mac]["curRx"].strip()
+                    value = float(raw) if raw != "" else 0.0
+                    monitor_devices[mac]["connection"]["speed"]["Rx"] = value
+                    raw = data[mac]["curTx"].strip()
+                    value = float(raw) if raw != "" else 0.0
+                    monitor_devices[mac]["connection"]["speed"]["Tx"] = value
+                
+                # Other info
+                if (data[mac]["vendor"] is not None
+                    and data[mac]["vendor"] != ""):
+                    monitor_devices[mac]["vendor"] = data[mac]["vendor"]
+
+        self._monitor_devices = monitor_devices
+
+        self._monitor_devices_time = now
+        self._monitor_devices_running = False
+
+        return
+
+
+    async def async_handle_error(self) -> None:
+        """Actions to be taken on connection errors"""
+
+        # Most of errors come from device reboots. Handle it
+        await self.async_handle_reboot()
+
+        # Clear error
+        await self.connection.async_reset_error()
+
+        return
+
+
+    async def async_handle_reboot(self) -> None:
+        """Actions to be taken on reboot"""
+
+        # Clear main monitor to prevent calculation errors in it
+        if self._monitor_main:
+            self._monitor_main = None
+            self._monitor_main_time = None
+
+        return
+
+
     async def async_find_interfaces(self, use_cache : bool = True) -> None:
         """Find available interfaces/type dictionary"""
         
@@ -632,10 +788,25 @@ class AsusRouter:
         await self.async_monitor_nvram()
         await self.async_find_interfaces()
         await self.async_monitor_misc()
+        await self.async_monitor_devices()
 
 
 #### RETURN DATA
 
+    async def async_get_devices(self, use_cache : bool = True) -> dict:
+        """Return device list"""
+        
+        now = datetime.utcnow()
+        if (self._monitor_devices is None
+            or self._monitor_devices_time is None
+            or use_cache == False
+            or (use_cache == True and self._monitor_devices_time and self._cache_time < (now - self._monitor_devices_time).total_seconds())
+        ):
+            await self.async_monitor_devices()
+
+        return self._monitor_devices
+
+
     async def async_get_cpu(self, use_cache : bool = True) -> dict:
         """Return CPU usage"""
         
@@ -722,21 +893,43 @@ class AsusRouter:
 
         return result
 
-    
+
     async def async_get_network_labels(self) -> list:
         """Return list of network interfaces"""
 
         if self._monitor_main is None:
             await self.async_monitor_main()
+        if self._monitor_nvram is None:
+            await self.async_monitor_nvram()
 
         result = list()
 
         for interface in self._monitor_main["NETWORK"]:
             result.append(interface)
 
+        if (not "USB" in result
+            and "dualwan" in self._monitor_nvram["rc_support"]
+        ):
+            result.append("USB")
+
         return result
 
 
+    async def async_get_ports(self, use_cache : bool = True) -> dict:
+        """Return WAN/LAN ports status"""
+        
+        now = datetime.utcnow()
+        if (self._monitor_misc is None
+            or use_cache == False
+            or (use_cache == True
+            and self._monitor_misc_time
+            and self._cache_time < (now - self._monitor_misc_time).total_seconds())
+        ):
+            await self.async_monitor_misc()
+
+        return self._monitor_misc["PORTS"]
+
+
     @property
     def connected(self) -> bool:
         return self.connection.connected
